import entropynodes.library as expNodes
from entropylab.flame.workflow import Workflow

wf = Workflow("Workflow name", description="What this realises in more details")

# now start workflow description

# ...


# end of workflow description

wf.register()
